"""
HTML Log Handler

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging

from ..format.command import indent_output

from ..provider.winrm import WinRM
from ..provider.wsmancli import WSManCLI

from ..response.fault import Fault
from ..response.instance import Instance
from ..response.reference import Reference
from ..response.association import Association


def html_render(value):
    return value.replace('<', '&#60;').replace('>', '&#62;').replace('\r\n', '\r\n<br/>').replace('\n', '\n<br/>').replace('\t','&nbsp;' * 8)


class HTMLHandler(logging.FileHandler):
    """
    Log handler that generates html command logs
    
    This handler only logs records that have a command, duration, and output property
    """
    
    def __init__(self, filename, encoding=None, delay=0, title="Command Log", pretty=False):
        """
        The constructor
        
        @param pretty: Determines if the output is formatted in a readable format or in raw XML
        @type pretty: boolean
        """
        self.title = title
        self.pretty = pretty
        
        # load the template files
        self.__preamble = open(os.path.dirname(__file__) + os.sep + 'templates' + os.sep + 'preamble.html').read()
        self.__tail = open(os.path.dirname(__file__) + os.sep + 'templates' + os.sep + 'tail.html').read()
        self.__command_template = open(os.path.dirname(__file__) + os.sep + 'templates' + os.sep + 'command.html').read()
        
        logging.FileHandler.__init__(self, filename, 'w', encoding, delay)
    
    def _open(self):
        """
        Internal open of the stream
        """
        stream = logging.FileHandler._open(self)
        log = self.__preamble % self.title
        log = log.replace('%%','%')
        stream.write(log)
        stream.write(self.__tail)
        stream.flush()
        return stream
    
    
    def emit(self, record):
        """
        Emit the message
        """
        
        # Only handle commands
        try:
            hasCommand = record.command
        except:
            hasCommand = False
        
        if hasCommand:
            
            # Set the pointer to the end of the command section
            if self.stream is None:
                self.stream = self._open()
            self.stream.seek(-len(self.__tail),2)
            
            if record.command.split(' ')[0] == 'winrm':
                provider = WinRM(None)
            else:
                provider = WSManCLI(None)
            
            # cast all responses to an iterable
            response = provider.parse(record.output)
            extra_class = "pass"
            if isinstance(response, Fault):
                extra_class = "fail"
                
            # generate the output
            if not self.pretty:
                entry = self.__command_template % (str(record.command), record.duration, extra_class, html_render(record.output))
            else:
                output =  indent_output(record)
                entry = self.__command_template % (str(record.command), record.duration, extra_class, html_render(output))
            
            # put out the message    
            self.stream.write(entry)
            self.stream.write(self.__tail)
            self.stream.flush()
            


"""
Association class

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.


from . import KeyResponse
from reference import Reference
from ..mixin.dictionary import DictionaryMixin


class Association(KeyResponse, DictionaryMixin):
    """
    Association class
    """
    
    def __init__(self, name):
        """
        Constructor for the association class
        
        @param name: Name of the instance
        @type name: String
        """
        
        # Call the base class
        super(Association, self).__init__()
        
        # Name of the reference 
        self.__name = name
    
    
    def toString(self, indent=0):
        s = [("\t" * indent) + self.name]
        
        for (k,v) in self.items:
            if isinstance(v[0], Reference):
                s.append(v[0].toString(indent + 1))
            else:
                s.append(("\t" * indent) + "\t%s=%s" % (k, ",".join(map(lambda x: x if x else "" ,v)) ))
        return "\n".join(s)
    
    def get_left_reference(self):
        for k,v in self.items:
            if isinstance(v,list) and len(v):
                ref = v[0]
                if isinstance(ref, Reference):
                    return (k,ref)
        return None
    
    def get_right_reference(self):
        foundLeft = False
        for k,v in self.items:
            if isinstance(v,list) and len(v):
                ref = v[0]
                if isinstance(ref, Reference):
                    if foundLeft:
                        return (k,ref)
                    else:
                        foundLeft = True
        return None
    
    def is_equal(self, assoc):
        """
        Tests this association instance for equalisty with the given associator
        
        @param assoc: The association class to compane with
        @type assoc: L{Association}
        """
        
        if isinstance(assoc, Association):
            for my_key, my_value in self.items:
                my_value = my_value[0] if my_value else None
                other_value = assoc.get(my_key, None)
                
                other_value = other_value[0] if other_value else None
                
                if isinstance(my_value, Reference):
                    if not isinstance(other_value, Reference):
                        # Cannot compare mixed types
                        print "Wrong type for key ", my_key, " expected Reference" 
                        return False
                    else:
                        # Compare the references
                        if not my_value.is_equal(other_value):
                            print "References are not equal"
                            return False
                elif my_value != other_value:
                    print "Key ", my_key, " not equal", my_value, " - ", other_value
                    return False
            return True
        else:
            print "Wrong type: expecting Association for comparison"
            return False
        
    
    # Properties of this class    
    name = property(fget=lambda x: x.__name)
    left_reference = property(fget=get_left_reference)
    right_reference = property(fget=get_right_reference)
     


"""
Reference class

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

from . import Response
from ..mixin.dictionary import DictionaryMixin


class Reference(Response, DictionaryMixin):
    """
    Reference object for WSMan
    """
    
    def __init__(self, name):
        """
        Constructor for the Reference class
        
        @param name: Name of the instance
        @type name: String
        """
        
        # Call the base class
        super(Reference, self).__init__()
        
        # Name of the reference 
        self.__name = name
        
        # Resource URI
        self.__resource_uri = ''
    
    def toString(self, indent=0):
        s = [("\t" * indent) + self.name]
        s.append(("\t" * indent) + "\tReferenceParameters")
        s.append(("\t" * indent) + "\t\tResourceURI=%s" % self.resource_uri)
        s.append(("\t" * indent) + "\t\tSelectorSet")
        for k,v in self.items:
            s.append(("\t" * indent) + "\t\t\tSelector: %s=%s" % (k, ",".join(map(lambda x: x if x else "" ,v)) ))
            
        return "\n".join(s)
    
    def set_resource_uri(self, uri):
        """
        Set the resource URI for the reference response
        """
        
        self.__resource_uri = uri
        
        
    def get_class_from_uri(self):
        """
        Get the class name of this reference from the URI
        """
        
        if self.__resource_uri:
            return self.__resource_uri.split('/')[-1]
        
        return ''            
    
    def is_equal(self, reference):
        """
        Tests this association instance for equalisty with the given associator
        
        @param reference: The association class to compane with
        @type reference: L{Reference}
        """
        
        if isinstance(reference, Reference):
            for my_key, my_value in self.items:
                other_value = reference.get(my_key, None)
                if other_value != my_value:
                    print "Reference Key ", my_key, " not equal", my_value, " - ", other_value
                    return False
                else:
                    print "Reference Key ", my_key, " equal", my_value, " - ", other_value
            return True
        else:
            print "Wrong type: expecting Reference for comparison"
            return False
        
    def __repr__(self):
        """
        String representation of the object
        """
        pairs = []
        for (key, value) in self.items:
            # Assemble the intermediate value since it is stored as a list
            # TODO: What if the value is an array? How do we assemble an array for a get query?
            value_ = value[0] if value else ''
            value_ = value_ if value_ else ''
            
            # JCT: used to wrap values with paces in double quotes - now we wrap the entire uri
            safe_value = value_
            pairs.append('%s=\"%s\"' % (key, safe_value))
        
        # Get the query parameter for command construction
        query = '&'.join(pairs)
        return "%s?%s" % (self.resource_uri, query)
            
            
    def __cmp__(self, other):
        """
        Comparison function
        """
        if self.resource_uri != other.resource_uri:
            return 1
        
        for (key, value) in other.items:
            if not self.has_key(key) or value != self.get(key):
                return -1
        return 0

    def __hash__(self):
        """
        The hash function
        """
        return hash("%s" % self)
        
    # Properties of this class    
    name = property(fget=lambda x: x.__name)
    classname = property(fget=get_class_from_uri)
    resource_uri = property(fget=lambda x: x.__resource_uri)
        
        


"""
WSManAPI

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import cache

from transport.process import Subprocess
from provider import WSManProviderFactory


class WSMan(object):
    """WS-management class"""
    
    def __init__(self, transport=Subprocess()):
        """
        Constructor for the WSMan class.
        
        @param transport: The L{transport} instance that will handle WSMan requests.  (default=L{Subprocess}) 
        @type transport: L{transport} 
        """
        
        # Store the transport
        self.__transport = transport
        
        # Provider
        self.__provider = WSManProviderFactory(self.__transport).get_provider()
    
    
    def identify(self, remote=None, raw=False):
        """
        Identify WS-Man implementation
        
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        @return: L{Response} object or the raw XML response
        @rtype: L{Response}
        """
        return self.__provider.identify(remote, raw)
    
    @cache.lru_cache(maxsize=20)    
    def enumerate(self, cim_class, cim_namespace, remote=None, raw=False, uri_host="http://schemas.dmtf.org"):
        """
        Enumerate a CIM class.
        
        @attention: Uses LRU Cache - set keyword argument I{cache} to "I{no}", "I{false}", or "I{off}" to bypass the cache.
        
        @param cim_class: CIM class to be enumerated
        @type cim_class: String
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: A list of L{Instance} objects or the raw XML response
        @rtype: list or string
        """
        
        return self.__provider.enumerate(cim_class, cim_namespace, remote, raw, uri_host)
    
    @cache.lru_cache(maxsize=20)
    def enumerate_keys(self, cim_class, cim_namespace, remote=None, raw=False, uri_host="http://schemas.dmtf.org"):
        """        
        Enumerate the keys for a CIM class.
        
        @attention: Uses LRU Cache - set keyword argument I{cache} to "I{no}", "I{false}", or "I{off}" to bypass the cache.
        
        @param cim_class: CIM class for key enumeration
        @type cim_class: String
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: A list of L{Instance} objects or the raw XML response
        @rtype: list or string
        """
        
        return self.__provider.enumerate_keys(cim_class, cim_namespace, remote, raw, uri_host)
    
    @cache.lru_cache(maxsize=20)
    def associators(self, instance, cim_namespace, remote=None, raw=False, uri_host="http://schemas.dmtf.org"):
        """
        Do an associators operation for the instance
        
        @attention: Uses LRU Cache - set keyword argument I{cache} to "I{no}", "I{false}", or "I{off}" to bypass the cache.
        
        @param instance: CIM instance response object
        @type instance: L{Instance}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: A list of L{Reference} objects or the raw XML response
        @rtype: list or string         
        """
                
        return self.__provider.associators(instance, cim_namespace, remote, raw, uri_host)
    
    @cache.lru_cache(maxsize=20)
    def references(self, instance, cim_namespace, remote=None, raw=False, uri_host="http://schemas.dmtf.org"):
        """
        Do a references operation for the instance
        
        @attention: Uses LRU Cache - set keyword argument I{cache} to "I{no}", "I{false}", or "I{off}" to bypass the cache.
        
        @param instance: CIM instance response object
        @type instance: L{Instance}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: A list of L{Reference} objects or the raw XML response
        @rtype: list or string         
        """
                
        return self.__provider.references(instance, cim_namespace, remote, raw, uri_host)
    
    def set(self, reference, cim_namespace, remote=None, properties={}, raw=False):
        """
        Sets the properties of an instance using the properties argument.
        
        @param reference: CIM reference response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param properties: The properties and values to set
        @type properties: dict 
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        
        @return: L{Response} object or the raw XML response
        @rtype: L{Response}
        """
        return self.__provider.set(reference, cim_namespace, remote, properties, raw)
    
    def get(self, reference, cim_namespace, remote=None, raw=False):
        """
        Do a get operation for the instance
        
        @param reference: CIM reference response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        
        @return: L{Instance} object or the raw XML response
        @rtype: L{Instance}         
        """
                
        return self.__provider.get(reference, cim_namespace, remote, raw)
    
    
    def invoke(self, reference, command, arguments, remote=None, raw=False):
        """
        Do a get operation for the instance
        
        @param reference: CIM reference response object
        @type reference: L{Reference}
        @param command: The command to invoke
        @type command: String
        @param arguments: A path to a file, or a dictionary of key/value paris used in the command
        @type arguments: String/dict
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        
        @return: L{Response} object or the raw XML response
        @rtype: L{Response}    
        """
        
        return self.__provider.invoke(reference, command, arguments, remote, raw)
    
    def __set_quiet(self, value):
        """
        Sets the transport's verbosity
        """
        self.__transport.quiet = value
    
    # Property to control the verbosity of the transport
    quiet = property(fset=__set_quiet)
        


"""functools.py - Tools for working with functions and callable objects
"""
# Python module wrapper for _functools C module
# to allow utilities written in Python to be added
# to the functools module.
# Written by Nick Coghlan <ncoghlan at gmail.com>
# and Raymond Hettinger <python at rcn.com>
#   Copyright (C) 2006-2010 Python Software Foundation.
# See C source code for _functools credits/copyright

"""
Modified to support cache bypass argument for decorated methods

@copyright: 2011
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

import sys

__all__ = ['update_wrapper', 'wraps', 'WRAPPER_ASSIGNMENTS', 'WRAPPER_UPDATES',
           'total_ordering', 'cmp_to_key', 'lru_cache', 'reduce', 'partial']

from _functools import partial, reduce
from collections import namedtuple
from ordereddict import OrderedDict

try:
    from thread import allocate_lock as Lock
except:
    print "ERROR importing _threa", sys.exc_info()
    from _dummy_thread import allocate_lock as Lock

# update_wrapper() and wraps() are tools to help write
# wrapper functions that can handle naive introspection

WRAPPER_ASSIGNMENTS = ('__module__', '__name__', '__doc__', '__annotations__')
WRAPPER_UPDATES = ('__dict__',)
def update_wrapper(wrapper,
                   wrapped,
                   assigned = WRAPPER_ASSIGNMENTS,
                   updated = WRAPPER_UPDATES):
    """Update a wrapper function to look like the wrapped function

       wrapper is the function to be updated
       wrapped is the original function
       assigned is a tuple naming the attributes assigned directly
       from the wrapped function to the wrapper function (defaults to
       functools.WRAPPER_ASSIGNMENTS)
       updated is a tuple naming the attributes of the wrapper that
       are updated with the corresponding attribute from the wrapped
       function (defaults to functools.WRAPPER_UPDATES)
    """
    wrapper.__wrapped__ = wrapped
    for attr in assigned:
        try:
            value = getattr(wrapped, attr)
        except AttributeError:
            pass
        else:
            setattr(wrapper, attr, value)
    for attr in updated:
        getattr(wrapper, attr).update(getattr(wrapped, attr, {}))
    # Return the wrapper so this can be used as a decorator via partial()
    return wrapper

def wraps(wrapped,
          assigned = WRAPPER_ASSIGNMENTS,
          updated = WRAPPER_UPDATES):
    """Decorator factory to apply update_wrapper() to a wrapper function

       Returns a decorator that invokes update_wrapper() with the decorated
       function as the wrapper argument and the arguments to wraps() as the
       remaining arguments. Default arguments are as for update_wrapper().
       This is a convenience function to simplify applying partial() to
       update_wrapper().
    """
    return partial(update_wrapper, wrapped=wrapped,
                   assigned=assigned, updated=updated)

def total_ordering(cls):
    """Class decorator that fills in missing ordering methods"""
    convert = {
        '__lt__': [('__gt__', lambda self, other: not (self < other or self == other)),
                   ('__le__', lambda self, other: self < other or self == other),
                   ('__ge__', lambda self, other: not self < other)],
        '__le__': [('__ge__', lambda self, other: not self <= other or self == other),
                   ('__lt__', lambda self, other: self <= other and not self == other),
                   ('__gt__', lambda self, other: not self <= other)],
        '__gt__': [('__lt__', lambda self, other: not (self > other or self == other)),
                   ('__ge__', lambda self, other: self > other or self == other),
                   ('__le__', lambda self, other: not self > other)],
        '__ge__': [('__le__', lambda self, other: (not self >= other) or self == other),
                   ('__gt__', lambda self, other: self >= other and not self == other),
                   ('__lt__', lambda self, other: not self >= other)]
    }
    # Find user-defined comparisons (not those inherited from object).
    roots = [op for op in convert if getattr(cls, op, None) is not getattr(object, op, None)]
    if not roots:
        raise ValueError('must define at least one ordering operation: < > <= >=')
    root = max(roots)       # prefer __lt__ to __le__ to __gt__ to __ge__
    for opname, opfunc in convert[root]:
        if opname not in roots:
            opfunc.__name__ = opname
            opfunc.__doc__ = getattr(int, opname).__doc__
            setattr(cls, opname, opfunc)
    return cls

def cmp_to_key(mycmp):
    """Convert a cmp= function into a key= function"""
    class K(object):
        __slots__ = ['obj']
        def __init__(self, obj, *args):
            self.obj = obj
        def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0
        def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0
        def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0
        def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0
        def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0
        def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0
        def __hash__(self):
            raise TypeError('hash not implemented')
    return K

_CacheInfo = namedtuple("CacheInfo", "hits misses maxsize currsize")

def lru_cache(maxsize=100):
    """Least-recently-used cache decorator.

    If *maxsize* is set to None, the LRU features are disabled and the cache
    can grow without bound.

    Arguments to the cached function must be hashable.

    View the cache statistics named tuple (hits, misses, maxsize, currsize) with
    f.cache_info().  Clear the cache and statistics with f.cache_clear().
    Access the underlying function with f.__wrapped__.

    See:  http://en.wikipedia.org/wiki/Cache_algorithms#Least_Recently_Used

    """
    # Users should only access the lru_cache through its public API:
    #       cache_info, cache_clear, and f.__wrapped__
    # The internals of the lru_cache are encapsulated for thread safety and
    # to allow the implementation to change (including a possible C version).

    def decorating_function(user_function,
                tuple=tuple, sorted=sorted, len=len, KeyError=KeyError):
        cache_stats = {'hits':0, 'misses': 0}
        kwd_mark = (object(),)          # separates positional and keyword args
        lock = Lock()                   # needed because ordereddicts aren't threadsafe

        
        cache = OrderedDict()       # ordered least recent to most recent
        cache_popitem = cache.popitem
        cache_renew = cache.move_to_end


        @wraps(user_function)
        def wrapper(*args, **kwds):
            # check for bypass
            use_cache = "True"
            as_tuple = "False"
            from_cache = False
            
            if kwds.has_key("cache"):
                try:
                    use_cache = kwds.pop("cache").lower()
                except:
                    pass
            
            if kwds.has_key("as_tuple"):
                try:
                    as_tuple = kwds.pop("as_tuple").lower()
                except:
                    pass
            
            
            #generate key
            key = args
            if kwds:
                key += kwd_mark + tuple(sorted(kwds.items()))
            
            # bypass cache?
            if use_cache.lower() == "false" or use_cache.lower() == "no" or use_cache.lower() == "off":
                result = user_function(*args, **kwds)
                with lock:
                    cache[key] = result     # record recent use of this key
                    cache_stats['misses'] += 1
                    if len(cache) > maxsize:
                        cache_popitem(0)    # purge least recently used cache entry
                if as_tuple.lower() == "true":
                    return (from_cache, "command", result)
                else:
                    return result
            
            try:
                with lock:
                    result = cache[key]
                    cache_renew(key)        # record recent use of this key
                    cache_stats['hits'] += 1
                    from_cache = True
            except KeyError:
                result = user_function(*args, **kwds)
                with lock:
                    cache[key] = result     # record recent use of this key
                    cache_stats['misses'] += 1
                    if len(cache) > maxsize:
                        cache_popitem(0)    # purge least recently used cache entry
            
            if as_tuple.lower() == "true":
                return (from_cache, "command", result)
            else:
                return result

        def cache_info():
            """Report cache statistics"""
            with lock:
                return _CacheInfo(cache_stats['hits'], cache_stats['misses'], maxsize, len(cache))

        def cache_clear():
            """Clear the cache and cache statistics"""
            #nonlocal hits, misses
            with lock:
                cache.clear()
                cache_stats['hits'] = 0
                cache_stats['misses'] = 0

        
        
        
        wrapper.cache_info = cache_info
        wrapper.cache_clear = cache_clear
        return wrapper

    # Added to help epydoc use the proper parameters for lru_cache decorated methods.
    try:
        decorating_function.__doc__ = user_function.__doc__
        decorating_function.__repr__ = user_function.__repr__
    except:
        pass
    return decorating_function



"""
Transport for WS-Management

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.


class Transport(object):
    """
    The Transport class is responsible for executing the
    constructed command.    
    """
    
    def __init__(self, **kwargs):
        """
        Constructor for the transport
        """
        
        # controls if the transport should write anything to the console
        self.__quiet = False
        pass
    
    
    def execute(self, command):
        """
        Execute the command and return the output.
        
        @param command: The command constructed by the provider.
        @type command: String
        
        @return: The output from the command execution 
        @rtype: String
        """
        
        raise NotImplementedError("This method needs to be implemented by the derived class.")
    
    def __set_quiet_mode(self, value):
        """
        Set verbosity of the transport
        """
        
        if isinstance(value, bool):
            self.__quiet = value
        else:
            pass
        
    
    quiet = property(fget=lambda x: x.__quiet, fset=__set_quiet_mode)    
        
    
    


"""
Parser class for XML to dictionary

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

import xml.parsers.expat as expat

class Parser(object):
    """
    XML to dictionary Parser
    """
    
    def __init__(self):
        """
        Constructor for the parser
        """
        
        # Stack for the elements
        self.__stack = []
        
        # Current element
        self.__current = None
        
        
    def reset(self):
        """
        Reset the markers 
        """
        
        self.__stack = []
        self.__current = None
        
        
    def element(self):
        """
        Get a new element for a node
        
        @return: Dictionary representation of the XML Node
        @rtype: Dictionary
        """
        
        return {'type':'', 'name':'', 'value':None, 'children':[], 'attributes':{}}
        
        
    def start(self, name, attributes):
        """
        Start tag handler for the node
        
        @param name: Name of the XML Node
        @type name: String
        @param attributes: Attributes of the XML Node
        @type attributes: Dictionary
        """
        
        # Get the qualifier and the name from the tag
        qname, tname = name.split(':')
        
        # If there is a current element then push it onto the stack
        if self.__current:
            self.__stack.append(self.__current)
        
        # Create a new element for this Node and set the properties
        self.__current               = self.element()
        self.__current['type']       = qname
        self.__current['name']       = tname
        self.__current['attributes'] = attributes
                    
    
    def character(self, data):
        """
        Character data handler of the XML node
        
        @param data: Node data
        @type data: String
        """
        
        if self.__current:
            
            # If the value is none, make it a string so we can append
            # - it's buffered data
            if not self.__current['value']:
                self.__current['value'] = ""
            
            self.__current['value'] += data
        
        
    def end(self, name):
        """
        End tag handler for the XML node
        """
        
        # Get the qualifier and the name from the tag
        qname, tname = name.split(':')
        
        # Get the parent if it exists
        parent = self.__stack.pop() if len(self.__stack) > 0 else None
        
        # Add it to the parent if it exists    
        if parent:
            parent['children'].append(self.__current)
            self.__current = parent
        
        
    def parse(self, xml):
        """
        Parse an XML string and return a dictionary representation.
        
        @param xml: XML string
        @type xml: String
        
        @return: Dictionary representation of the XML
        @rtype: Dictionary 
        """
        
        # Reset the stack and the current element
        self.reset()
        
        # Create the parser
        parser = expat.ParserCreate()
        
        parser.buffer_text = True
        
        # Set the handlers
        parser.StartElementHandler  = self.start
        parser.CharacterDataHandler = self.character
        parser.EndElementHandler    = self.end
        
        # Parse the XML        
        parser.Parse(xml)
        
        return self.__current
    


"""
Dictionary Mix-in

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

from pprint import pformat

class DictionaryMixin(object):
    """
    Dictionary Mixin class
    """
    
    def __init__(self):
        """
        Constructor for the Dictionary Mixin
        """
        
        super(DictionaryMixin, self).__init__()
        
        # Dictionary
        self.__mapping = {}
        
        # Case insensitive dictionary
        self.__lower_mapping = {}
        
        
    def set(self, key, value):
        """
        Add/Edit a dictionary mapping
        
        Example Output:
        <..>
        <n1:IdentifyingDescriptions>CIM:GUID</n1:IdentifyingDescriptions>
        <n1:IdentifyingDescriptions>CIM:Tag</n1:IdentifyingDescriptions>
        <n1:IdentifyingDescriptions>DCIM:ServiceTag</n1:IdentifyingDescriptions>
        <..>
        
        IdentifyingDescriptions is an array cause it is repeated so we need to store
        all of its values hence the value for a key is stored as a list
        
        @param key: Key for the mapping
        @type key: String
        @param value: Value of the mapping
        @type value: String                 
        """
        
        if self.__mapping.has_key(key):
            self.__mapping[key].append(value)
            self.__lower_mapping[key.lower()].append(value)
        else:
            self.__mapping[key] = [value]
            self.__lower_mapping[key.lower()] = [value]
                 
        return True
        
        
    def get(self, key, default=None):
        """
        Get the value of the key from the mapping
        
        @param key: Key for the mapping
        @type key: String
        @param default: Default value in case the mapping does not exist
        @type default: L{object}
        
        @return: The value of the mapping if it exists else default.
        @rtype: String or L{object}        
        """
        
        #return self.__mapping.get(key, default)
        return self.__lower_mapping.get(key.lower(), default)
    
    def has_key(self, key):
        """
        Checks if a key is set/defined
        
        @param key: Key for the mapping
        @type key: String
        
        @return: True if defined
        @rtype: boolean
        """
        
        #return self.__mapping.has_key(key)
        return self.__lower_mapping.has_key(key.lower())
    
    def dump(self):
        try:
            return pformat(self.__mapping)
        except:
            return str(self)    
        
    # Properties 
    keys    = property(fget=lambda x: x.__mapping.keys()) 
    values  = property(fget=lambda x: x.__mapping.values())
    items   = property(fget=lambda x: x.__mapping.items())
    


"""
Fault response

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.


from . import Response

class Fault(Response):
    """
    Fault response
    """
    
    def __init__(self, code, reason, detail):
        """
        Constructor for the Fault response
        
        @param code: Fault code
        @type code: String
        @param reason: Reason for the fault
        @type reason: String
        @param detail: URL for the fault details
        @type detail: String
        """
        
        # Call the base class
        super(Fault, self).__init__()
        
        # Fault attributes
        self.__code     = code
        self.__reason   = reason
        self.__detail   = detail
    
    def toString(self):
        """
        show the object in a reader friendly format
        """
        s = "Fault:\n\tCode: %s\n\tReason: %s\n\tDetail: %s" % (self.code, self.reason, self.detail)
        return s
        
        
        
        
    # Fault properties
    code    = property(fget=lambda x: x.__code)
    reason  = property(fget=lambda x: x.__reason)
    detail  = property(fget=lambda x: x.__detail)
    


"""
WSMan Command Log Handler

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

import sys
import logging

from ..provider.winrm import WinRM
from ..provider.wsmancli import WSManCLI

from ..response.fault import Fault
from ..response.instance import Instance
from ..response.reference import Reference
from ..response.association import Association


def indent_output(record):
    """
    Formats command records based on the command
    
    @param record: A log record object that must have a command and output property
    
    @returns: String
    """
    if record.command.split(' ')[0] == 'winrm':
        provider = WinRM(None)
    else:
        provider = WSManCLI(None)
    
    # cast all responses to an iterable
    response = provider.parse(record.output)
    if not isinstance(response,list):
        response = [response]
    
    # build the string
    strs = []
    for i in response:
        strs.append(i.toString())
    return '\n'.join(strs)
        
    
    

class OutputFormatter(logging.Formatter):
    """
    Formats WSMan command output suitable for use with 
    logging.Handlers
    """
    
    def __init__(self, fmt=None, dateFmt=None, pretty=False):
        """
        Constructor
        
        @param pretty: Set to true to get indented output
        """
        self.pretty = pretty
        logging.Formatter.__init__(self, fmt, dateFmt)
        
        
    def format(self, record):
        """
        Formats the record
        """
        
        try:
            output = record.output
        except:
            output = False    
            
        if output:
            if self.pretty:
                command = record.command
                record.output = indent_output(record) + '\n'
                record.command = '\n' + record.command + '\n'
                msg = logging.Formatter.format(self, record)
                record.output = output
                record.command = command
                return msg
            else:
                command = record.command
                record.command = '\n' + record.command + '\n'
                msg = logging.Formatter.format(self, record)
                record.command = command
                return msg
        else:
            record.command = ""
            record.output = ""
            record.duration = 0.0
            msg = logging.Formatter.format(self, record)
            del record.command
            del record.output
            del record.duration
            return msg
            
     
     
        
        


"""
Remote configuration

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

class Remote(object):
    """
    Remote configuration class
    """
    
    def __init__(self, ip, username, password):
        """
        Constructor for the remote configuration
        
        @param ip: IP address of the SUT
        @type ip: String
        @param username: Username for the SUT
        @type username: String
        @param password: Password for the SUT
        @type password: String  
        """
        
        #Store the IP, Username and Password
        self.__ip = ip
        self.__username = username
        self.__password = password
    
    
    def __cmp__(self, other):
        if self.__ip == other.ip and self.__password == other.password and self.__username == other.username:
            return 0
        else:
            return 1
        
    def __hash__(self):
        """
        The hash function
        """
        return hash("%s %s %s" % (self.__ip, self.__username, self.__password))
        
            
    # Properties of the Remote class
    ip       = property(fget=lambda x: x.__ip)
    username = property(fget=lambda x: x.__username)
    password = property(fget=lambda x: x.__password)
    
    
if __name__ == "__main__":
    
    remote = Remote('127.0.0.1', 'root', 'calvin')
    print remote.ip, remote.username, remote.password
    try:
        remote.ip = '192.168.0.1'
    except AttributeError:        
        print "Check: This attribute cannot be set"


"""
Dummy transport 

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
import time
import random
import logging

log = logging.getLogger("WSMAN.transport")
from .. import Transport


class Dummy(Transport):
    """
    Dummy transport
    """
    
    def execute(self, command):
        """
        Execute the command and return the output.
        
        @param command: The command constructed by the provider.
        @type command: String
        
        @return: The output from the command execution 
        @rtype: String
        """
        start = time.time()
        
        # You could parse the command line here and send the appropriate file
        # but the parser needs to handle all of these formats. This transport
        # picks the file randomly. 
        
        files = ['wsmanfault.txt', 'epr.txt', 'get.txt', 'fault.txt', 'instance.txt', 'instances.txt', 'invoke_resp.txt', 'invoke_result.txt']                
        #files = ['invoke_resp.txt', 'invoke_result.txt']
        #files = ['get.txt', 'instance.txt', 'instances.txt', 'wsmanfault.txt']
        #files = ['fault.txt']
        files = ['large_response.txt']
        
        filename = random.choice(files)
        basename = os.path.dirname(__file__)
        #provider = 'winrm' if (sys.platform == 'win32') else 'wsmancli'
        provider = 'winrm'
        path = os.path.join(basename, 'responses', provider, filename)
        print path
        
        duration = time.time() - start
        output = open(path, "r").read()
        log.info("Command Completed in %0.3f s" % duration, extra={'command': command, 'output': output, 'duration':duration})
        return output
    



"""
WinRM based provider

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

from pprint import pprint

from wsman import WSManProvider
from ..parsers import Parser
from ..response.fault import Fault
from ..response.instance import Instance
from ..response.reference import Reference
from ..response.association import Association
from exceptions import TypeError

import logging
log = logging.getLogger("WSMAN")



class WinRM(WSManProvider):
    """
    Windows based WS-Management provider.
    """
    
    def __init__(self, transport):
        """
        Constructor for the WSMan provider class.
        
        @param transport: Transport object to execute the constructed command
        @type transport: L{Transport}
        """
        
        # Call the base class 
        super(WinRM, self).__init__(transport)
        
        
    def extract(self, output):
        """
        Extract the results XML or the fault XML from the output 
        
        @param output: Output from the transport
        @type output: String
        @return: Result XML from the output
        @rtype: String
        """
        
        def extract_(tag):
            start = output.find('<%s' % tag)
            if start > -1:
                end = output.find('</%s>' % tag)
                if end > -1:
                    xml = output[start:end] + '</%s>' % tag
                    return xml            
            return None
        
        # Extract the XML from the output
        xml = extract_('wsman:Results')
        xml = extract_('s:Fault') if not xml else xml
        xml = extract_('f:WSManFault') if not xml else xml
        xml = extract_('wsmid:IdentifyResponse') if not xml else xml
              
        return xml
    
    
    def response_from_reference(self, name, node):
        """
        Construct the reference object from the node
        
        @param name: Name of this reference
        @type name: String
        @param node: XML reference dictionary
        @type node: Dictionary
        
        @return: Response object
        @rtype: L{Response}
        """
        
        # Reference object
        reference = Reference(name)
        
        # Get the selectors
        for child in node.get('children', []):
            
            # Set the resource URI for the reference
            if child.get('name', '') == 'ResourceURI':
                reference.set_resource_uri(child.get('value', ''))
                
            # Set the selector set                                
            elif child.get('name', '') == 'SelectorSet':
                for child_ in child.get('children', []):
                    if child_.get('name', '') == 'Selector':
                        key   = child_.get('attributes', {}).get('Name', '')
                        value = child_.get('value', None)
                                                                        
                        if key:
                            reference.set(key, value)                        
        return reference
    
    
    def is_association(self, node):
        """
        Check to see if this is an association node
        
        @param node: XML results node
        @type node: Dictionary
        """
        
        # Check if any of the node children have children
        if node and isinstance(node, dict):
            for child in node.get('children', []):
                if child.get('children', []):
                    return True
        return False    


    def response_from_identify(self, node):
        """
        Construct the response object from the results node dictionary
        
        @param node: XML results dictionary
        @type node: Dictionary
        
        @return: Response object parsed from the node
        @rtype: L{Response}
        """
        
        name = node.get("name",'')
        response = Instance(name)
        for child_ in node.get('children', []):
                    
            # Get the name and value
            key = child_.get('name', None)
            value = child_.get('value', None)
            
            # Construct the association
            if key and child_.get('children', []):
                if key == 'SecurityProfiles':
                    for child__ in child_.get('children', []):
                        if child__.get('name', '') == 'SecurityProfileNames':
                            response.set(key,child__.get('value', ''))
                     
            elif key:
                response.set(key, value)
            
        return response
        
    
    def response_from_results(self, node):
        """
        Construct the response object from the results node dictionary
        
        @param node: XML results dictionary
        @type node: Dictionary
        
        @return: List of Response objects parsed from the node
        @rtype: List of L{Response}
        """
        
        # Response objects
        responses = []
        
        # Decide if this is an Instance, Association or Reference
        for child in node.get('children', []):
            qname = child.get('type', '')
            name = child.get('name', '')
            
            # Response Object
            response = None
            
            # End point reference (EPR)
            if name == 'EndpointReference': 
                for child_ in child.get('children', []):
                    if child_.get('name', '') == 'ReferenceParameters':
                        response = self.response_from_reference(name, child_)
                       
            # Instance or Association
            else:
                
                response = Instance(name) if not self.is_association(child) else Association(name)                
                #print "PARSING", name, response
                # Build the instance attributes
                for child_ in child.get('children', []):
                    
                    # Get the name and value
                    key = child_.get('name', None)
                    value = child_.get('value', None)
                    
                    #print "Processing", key
                    # Construct the association
                    if key and child_.get('children', []):                                                
                        for child__ in child_.get('children', []):
                            #print "\tChild", child__.get("name",""), "of", key
                            if child__.get('name', '') == 'ReferenceParameters':
                                value = self.response_from_reference(key, child__)
                                
                            # Handle invoke responses with a non-standard format (Modular 12G) 
                            elif child__.get('name', '') == 'EndpointReference':
                                try:
                                    value = self.response_from_results(child_)[0]
                                except:
                                    value = Reference(child__.get('name', ''))
                    if key:
                        nilSet=False
                        for aKey_, aValue_ in child_.get("attributes",{}).items():
                            aKey_ = aKey_.split(":")[-1]
                            if aKey_ == "nil" and aValue_ == "true":
                                response.set(key, None)
                                nilSet = True
                                break                                
                        if not nilSet:
                            response.set(key, value)
                                                
            if response:
                responses.append(response)
                
        return responses
    
    
    def response_from_fault(self, node):
        """
        Construct the response object from the fault node dictionary
        
        @param node: XML fault dictionary
        @type node: Dictionary
        
        @return: Response object
        @rtype: L{Response}
        """
                
        code   = 'WinRM'
        reason = 'Internal Server Error (WinRM Provider)'
        detail = 'Internal Server Error (WinRM Provider)'
        
        try:
            for child in node.get('children', []):
                name = child.get('name', '')
                if name == 'Code':
                    for x in child.get('children', []):
                        if x.get('name') == 'Subcode':
                            for y in x.get('children', []):
                                if y.get('name') == 'Value':
                                    code = y.get('value', '')
                                                                                                        
                elif name == 'Reason':
                    for x in child.get('children', []):
                        if x.get('name') == 'Text':                        
                            reason = x.get('value', '')
                                  
                elif name == 'Detail':
                    for x in child.get('children', []):
                        if x.get('name') == 'FaultDetail':                        
                            detail = x.get('value', '')                            
        except:
            pass
            
        return Fault(code, reason, detail)
    
    
    def response_from_wsmanfault(self, node):
        """
        Construct the response object from the wsman fault node dictionary
        
        @param node: XML fault dictionary
        @type node: Dictionary
        
        @return: Response object
        @rtype: L{Response}
        """
                
        code   = 'WinRM'
        reason = 'Internal Server Error (WinRM Provider)'
        detail = 'Internal Server Error (WinRM Provider)'
        
        try:
            for child in node.get('children', []):
                log.info("Fault node message %s" % child.get("name"))
                if child.get('name') == 'Message':
                    code = node.get('attributes', {}).get('Code', code)
                    reason = child.get('value', reason)
        except:
            log.error("Error getting fault %s:%s" %(sys.exc_info()[1],sys.exc_info()[0]))
            pass
            
        return Fault(code, reason, detail)
        
        
    def parse(self, output):
        """
        Parse the output into one of the response formats.
        
        @param output: Output from the transport
        @type output: String
        """
        # Extract the XML from the output
        xml = self.extract(output)
        
        # Get the  dictionary representation of the extracted XML
        xml_dict = Parser().parse(xml) if xml else {}
        
        # If it is a results XML
        if xml_dict.get('name', '') == 'Results':
            return self.response_from_results(xml_dict)
        
        elif xml_dict.get('name', '') == 'IdentifyResponse':
            return self.response_from_identify(xml_dict)
        
        # WSMan Fault
        elif xml_dict.get('name', '') == 'WSManFault':
            return  self.response_from_wsmanfault(xml_dict)
    
        elif xml_dict.get('name', '') == 'Fault':
            return  self.response_from_fault(xml_dict)
        
        # Fault!
        else:
            try:
                return self.response_from_results(Parser().parse("<wsman:Results>%s</wsman:Results>" % output) if output else {})
            except:
                pass
            return self.response_from_fault(xml_dict)
        
        
        
    def remote_options(self, remote):
        """
        Generate the options for a remote command.
        
        @param remote: Remote configuration object
        @type remote: L{Remote}
        
        @return: Remote command line options
        @rtype: String
        """
        
        if not remote:
            return ''
        
        remote_  = '-u:%s ' % remote.username
        remote_ += '-p:%s ' % remote.password
        remote_ += '-r:https://%s/wsman ' % remote.ip
        remote_ += '-encoding:utf-8 -a:basic '
        
        return remote_
    
    def properties_argument(self, properties):
        """
        Builds the set argument for the set command
        
        @param properties: the properties to be set
        @type properties: dict
        
        @return: A string used to set properties of an instance in the form @{k1="v1";kn="vn"}
        @rtype: String
        """
        kv_strs=[]
        for k,v in properties.items():
            kv_strs.append("%s=\"%s\"" % (k,v))
        
        kv_strs = ";".join(kv_strs)
           
        return "@{%s}" % kv_strs
    
    
    def identify(self, remote=None, raw=False):
        """
        Identify WS-Man implementation
        
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        @return: L{Response} object or the raw XML response
        @rtype: L{Response}
        """    
        # Construct the command
        command  = 'winrm id '
        command += self.remote_options(remote)        
        command += '-SkipCNcheck -SkipCAcheck -format:Pretty'
        
        # Use the transport and execute the command
        output = self.get_transport().execute(command)        
        if raw: 
            return output
        else:
            # Parse the output into a response object
            return self.parse(output)
    
        
    def enumerate(self, cim_class, cim_namespace, remote=None, raw=False, uri_host=""):
        """
        Enumerate the CIM class.
        
        @param cim_class: CIM class to be enumerated
        @type cim_class: String
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response objects after enumeration
        @rtype: List of L{Response} objects/ L{Fault}
        """
        
        # Construct the command
        enumerate_command  = 'winrm e \"%s/wbem/wscim/1/cim-schema/2/%s?__cimnamespace=%s\" ' % (uri_host, cim_class, cim_namespace)
        enumerate_command += self.remote_options(remote)        
        enumerate_command += '-SkipCNcheck -SkipCAcheck -format:Pretty'
        
        # Use the transport and execute the command
        output = self.get_transport().execute(enumerate_command)        
        if raw: 
            return output
        else:
            # Parse the output into a response object
            return self.parse(output)
    
    
    def enumerate_keys(self, cim_class, cim_namespace, remote=None, raw=False, uri_host=""):
        """
        Enumerate the keys for the cim class.
        
        @param cim_class: CIM class for key enumeration
        @type cim_class: String
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response objects after enumerating the keys
        @rtype: List of L{Reference} objects/ L{Fault}
        """
        
        # Construct the command
        enumerate_command  = 'winrm e \"%s/wbem/wscim/1/cim-schema/2/%s?__cimnamespace=%s\" ' % (uri_host, cim_class, cim_namespace)
        enumerate_command += self.remote_options(remote)
        enumerate_command += '-SkipCNcheck -SkipCAcheck -format:Pretty -ReturnType:EPR'
        
        # Use the transport and execute the command
        output = self.get_transport().execute(enumerate_command)
        
        if raw: 
            return output
        else:
            # Parse the output into a response object
            return self.parse(output)
    
    
    
    def set(self, reference, cim_namespace, remote=None, properties={}, raw=False):
        """
        Sets the properties of an instance using the properties argument.
        
        @param reference: CIM reference response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class (depricated: using the namespace from the reference)
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param properties: The properties and values to set
        @type properties: dict 
        
        @return: Instance object or Fault
        @rtype: L{Response}
        """
        
        # Get the keys reference for this instance (This needs to be set prior to the get)
        if reference and\
            isinstance(reference, Reference):
            
            # Important: 
            # Form the query from the reference and not from the enumerated instance
            pairs = []
            for (key, value) in reference.items:
                                        
                # Assemble the intermediate value since it is stored as a list
                # TODO: What if the value is an array? How do we assemble an array for a get query?
                value_ = value[0] if value else ''
                value_ = value_ if value_ else ''
                
                # JCT: used to wrap values with paces in double quotes - now we wrap the entire uri
                safe_value = value_
                pairs.append('%s=%s' % (key, safe_value))
            
            # Get the query parameter for command construction
            query = '+'.join(pairs)
            classname = reference.classname
            
            # Construct the command
            get_command  = 'winrm s \"%s?%s\" ' % (reference.resource_uri, query)
            get_command += self.properties_argument(properties) + ' '
            get_command += self.remote_options(remote)
            get_command += '-SkipCNcheck -SkipCAcheck -format:Pretty'
            
            # Use the transport and execute the command
            output = self.get_transport().execute(get_command)
    
            if raw:
                return output
            
            else:
                # Note: (WinRM specific actions)
                # If it is not a fault then add the covers since get does not return
                # the results within the <wsman:Results> tag.
                if not self.extract(output):
                    output = '<wsman:Results>' + output + '</wsman:Results>'
                
                
                # Parse the output into a response object
                instance_or_fault = self.parse(output)
                log.debug("Got a %s response" % (type(instance_or_fault)))
                if not isinstance(instance_or_fault, Fault) and isinstance(instance_or_fault, list): 
                    if len(instance_or_fault) > 0:
                        return instance_or_fault[0]
                    else:
                        return Fault('WinRM','Error', "Invalid response")
                        
                    
                return instance_or_fault
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
        
        # Generate a Fault
        return Fault('WinRM',\
                     'Set is not supported on this instance or the reference is not set or it returned an error.',\
                     'WinRM provider for WSMAN returned an error.')
    
        
    def associators(self, reference, cim_namespace, remote=None, raw=False, uri_host=""):
        """
        Do an associators operation for an instance
        
        @param reference: CIM reference(key) response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response objects after enumerating the keys
        @rtype: List of L{Reference} objects/ L{Fault}
        """
        # Get the keys reference for this instance (This needs to be set prior to the get)
        if reference and\
            isinstance(reference, Reference):
            
            pairs = []
            for (key, value) in reference.items:
                                        
                # Assemble the intermediate value since it is stored as a list
                # TODO: What if the value is an array? How do we assemble an array for a get query?
                value_ = value[0] if value else ''
                value_ = value_ if value_ else ''
                
                # JCT: used to wrap values with paces in double quotes - now we wrap the entire uri
                safe_value = value_
                pairs.append('%s=%s' % (key, safe_value))
        
            # Get the query parameter for command construction
            query = '+'.join(pairs)
            classname = reference.classname
            
            get_command = 'winrm e %s/wbem/wscim/1/cim-schema/2/* ' % (uri_host) 
            get_command += '-dialect:association -filter:{object=%s?%s} ' % (classname, query)
            get_command += self.remote_options(remote)
            get_command += '-SkipCNcheck -SkipCAcheck -format:Pretty'
            
            # Use the transport and execute the command
            output = self.get_transport().execute(get_command)
            
            if raw: 
                return output
            else:
                # Parse the output into a response object
                return self.parse(output)
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
        
        
        # Generate a Fault
        return Fault('WinRM',\
                 'Get is not supported on this instance or the reference is not set or it returned an error.',\
                 'WinRM provider for WSMAN returned an error.')
    
    def references(self, reference, cim_namespace, remote=None, raw=False, uri_host=""):
        """
        Do a references operation for an instance
        
        @param reference: CIM reference(key) response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response objects after enumerating the keys
        @rtype: List of L{Reference} objects/ L{Fault}
        """
        # Get the keys reference for this instance (This needs to be set prior to the get)
        if reference and\
            isinstance(reference, Reference):
            
            pairs = []
            for (key, value) in reference.items:
                                        
                # Assemble the intermediate value since it is stored as a list
                # TODO: What if the value is an array? How do we assemble an array for a get query?
                value_ = value[0] if value else ''
                value_ = value_ if value_ else ''
                
                # JCT: used to wrap values with paces in double quotes - now we wrap the entire uri
                safe_value = value_
                pairs.append('%s=%s' % (key, safe_value))
        
            # Get the query parameter for command construction
            query = '+'.join(pairs)
            classname = reference.classname
            
            get_command = 'winrm e %s/wbem/wscim/1/cim-schema/2/* ' % (uri_host)
            get_command += '-dialect:association -associations -filter:{object=%s?%s} ' % (classname, query)
            get_command += self.remote_options(remote)
            get_command += '-SkipCNcheck -SkipCAcheck -format:Pretty'
            
            # Use the transport and execute the command
            output = self.get_transport().execute(get_command)
            
            if raw: 
                return output
            else:
                # Parse the output into a response object
                return self.parse(output)
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
        
        
        # Generate a Fault
        return Fault('WinRM',\
                 'Get is not supported on this instance or the reference is not set or it returned an error.',\
                 'WinRM provider for WSMAN returned an error.')

    
            
    def get(self, reference, cim_namespace, remote=None, raw=False):
        """
        Do a get operation for a reference. 
        
        @param reference: CIM reference(key) response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        
        @return: Response object after enumerating the keys
        @rtype: L{Instance}         
        """
        
        # Get the keys reference for this instance (This needs to be set prior to the get)
        if reference and\
            isinstance(reference, Reference):
            
            # Important: 
            # Form the query from the reference and not from the enumerated instance
            pairs = []
            for (key, value) in reference.items:
                # Assemble the intermediate value since it is stored as a list
                # TODO: What if the value is an array? How do we assemble an array for a get query?
                value_ = value[0] if value else ''
                value_ = value_ if value_ else ''
                
                # JCT: used to wrap values with paces in double quotes - now we wrap the entire uri
                safe_value = value_
                pairs.append('%s=%s' % (key, safe_value))
        
            # Get the query parameter for command construction
            query = '+'.join(pairs)
            classname = reference.classname
            
            # Construct the command
            get_command  = 'winrm g \"%s?%s\" ' % (reference.resource_uri, query)
            get_command += self.remote_options(remote)
            get_command += '-SkipCNcheck -SkipCAcheck -format:Pretty'
            
            
            # Use the transport and execute the command
            output = self.get_transport().execute(get_command)
        
            if raw:
                return output
            
            else:
                # Note: (WinRM specific actions)
                # If it is not a fault then add the covers since get does not return
                # the results within the <wsman:Results> tag.
                if not self.extract(output):
                    output = '<wsman:Results>' + output + '</wsman:Results>'
                
                # Parse the output into a response object
                instance_or_fault = self.parse(output)
                if not isinstance(instance_or_fault, Fault) and isinstance(instance_or_fault, list):                
                    if len(instance_or_fault) > 0:
                        return instance_or_fault[0]
                    else:
                        return Fault('WinRM',\
                                     'Get is not supported on this instance or the reference is not set or it returned an error.',\
                                     'WinRM provider for WSMAN returned an error.')
                    
                return instance_or_fault
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
        
        # Generate a Fault
        return Fault('WinRM',\
                     'Get is not supported on this instance or the reference is not set or it returned an error.',\
                     'WinRM provider for WSMAN returned an error.')
        
    
    def invoke(self, reference, command, arguments, remote=None, raw=False):
        """
        Do a get operation for the instance
        
        @param reference: CIM reference response object
        @type reference: L{Reference}
        @param command: The command to invoke
        @type command: String
        @param arguments: A path to a file, or a dictionary of key/value paris used in the command
        @type arguments: String/dict
        @param remote: Remote configuration object
        @type remote: L{Remote}
        
        @return: Response object from the call, either a Reference, Fault, or an Result
        @rtype: L{Response}    
        """
        
        
        if reference and isinstance(reference, Reference):
            # Important: 
            # Form the query from the reference and not from the enumerated instance
            pairs = []
            
            for (key, value) in reference.items:
                # Assemble the intermediate value since it is stored as a list
                # TODO: What if the value is an array? How do we assemble an array for a get query?
                value_ = value[0] if value else ''
                value_ = value_ if value_ else ''
                
                # JCT: used to wrap values with paces in double quotes - now we wrap the entire uri
                safe_value = value_
                pairs.append('%s=%s' % (key, safe_value))
            
            # Get the query parameter for command construction
            query = '+'.join(pairs)
            classname = reference.classname
            
            # Construct the command
            get_command  = 'winrm i %s \"%s?%s\" ' % (command, reference.resource_uri, query)
            get_command += self.remote_options(remote)
            get_command += '-SkipCNcheck -SkipCAcheck -format:Pretty '
            
            # construct the options
            if isinstance(arguments, basestring):
                get_command += "-file:\"%s\"" % arguments
            elif isinstance(arguments, dict):
                get_command += " " + self.properties_argument(arguments)
             
            # Use the transport and execute the command
            output = self.get_transport().execute(get_command)
            
            if raw:
                return output
            else:
                # Parse the output into a response object
                instance_or_fault = self.parse(output)
                if not isinstance(instance_or_fault, Fault) and isinstance(instance_or_fault, list):                
                    if len(instance_or_fault) > 0:
                        return instance_or_fault[0]
                    else:
                        return Fault('WinRM',\
                                     'Invoke is not supported on this instance or the reference is not set or it returned an error.',\
                                     'WinRM provider for WSMAN returned an error.')
                        
                    
                return instance_or_fault
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
        
        # Generate a Fault
        return Fault('WinRM',\
                     'Invoke is not supported on this instance or the reference is not set or it returned an error.',\
                     'WinRM provider for WSMAN returned an error.')


"""
WSManCLI based provider

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

import sys

from wsman import WSManProvider

from ..parsers import Parser
from ..response.fault import Fault
from ..response.instance import Instance
from ..response.reference import Reference
from ..response.association import Association
from exceptions import TypeError

import logging
log = logging.getLogger("WSMAN")

def find_children(name, node):
    """
    Finds direct children of a node by name
    
    @param name: The name of the node to find
    @type name: string
    
    @param node: the dict node
    @type node: dict
    """
    
    nodes = []
    try:
        for child in node.get('children',[]) if node else []:
            if child.get('name','') == name:
                nodes.append(child)
    except:
        pass
    return nodes 


def find_child(name, node):
    """
    Finds a direct child by name
    
    @param name: The node name to find (aka the Tag)
    @type name: string
    
    @param node: The dict node
    @type node: dict
    """
    try:
        for child in node.get('children',[]) if node else []:
            if child.get('name','') == name:
                return child
    except:
        pass
    return None
            

class WSManCLI(WSManProvider):
    """
    Unix based WS-Management provider.
    """
    
    def __init__(self, transport):
        """
        Constructor for the WSMan provider class.
        
        @param transport: Transport object to execute the constructed command
        @type transport: L{Transport}
        """
        
        # Call the base class 
        super(WSManCLI, self).__init__(transport)
    
    def is_association(self, node):
        """
        Check to see if this is an association node
        
        @param node: XML results node
        @type node: Dictionary
        """
        
        # Check if any of the node children have children
        if node and isinstance(node, dict):
            for child in node.get('children', []):
                if child.get('children', []):
                    return True
        return False    
    

    
    def response_from_reference(self, name, node):
        """
        Construct the reference object from the node
        
        @param name: Name of this reference
        @type name: String
        @param node: XML reference dictionary
        @type node: Dictionary
        
        @return: Response object
        @rtype: L{Response}
        """
        
        # Reference object
        reference = Reference(name)
        
        # Get the selectors
        for child in node.get('children', []):
            
            # Set the resource URI for the reference
            if child.get('name', '') == 'ResourceURI':
                reference.set_resource_uri(child.get('value', ''))
                
            # Set the selector set                                
            elif child.get('name', '') == 'SelectorSet':
                for child_ in child.get('children', []):
                    if child_.get('name', '') == 'Selector':
                        key   = child_.get('attributes', {}).get('Name', '')
                        value = child_.get('value', None)
                                                                        
                        if key:
                            reference.set(key, value)                        
        return reference
    
    
    def response_from_identify(self, node):
        """
        Construct the response object from the results node dictionary
        
        @param node: XML results dictionary
        @type node: Dictionary
        
        @return: Response object parsed from the node
        @rtype: L{Response}
        """
        
        name = node.get("name",'')
        response = Instance(name)
        for child_ in node.get('children', []):
                    
            # Get the name and value
            key = child_.get('name', None)
            value = child_.get('value', None)
            
            # Construct the association
            if key and child_.get('children', []):
                if key == 'SecurityProfiles':
                    for child__ in child_.get('children', []):
                        if child__.get('name', '') == 'SecurityProfileNames':
                            response.set(key,child__.get('value', ''))
                     
            elif key:
                response.set(key, value)
            
        return response
    
    
    def generate_response(self, node):
        """
        Construct the response object from the results node dictionary
        
        @param node: XML results dictionary
        @type node: Dictionary
        
        @return: List of Response objects parsed from the node
        @rtype: List of L{Response}
        """
        item_nodes = []
        
        if not node:
            return Fault('WSManCLI', 
                         'Internal Server Error(WSManCLI Provider)', 
                         'Invalid response format')
            
        # Response objects
        responses = []
        
        fault = find_child("WSManFault", node)
        if fault:
            return self.response_from_fault(fault)
        
        # Iterate over the envelopes - may be more than one (see PullResponse)
        # to get the items
        envelopes = find_children("Envelope", node)
        if not envelopes:
            return Fault('WSManCLI', 
                         'Internal Server Error(WSManCLI Provider)', 
                         'Invalid response format - no Envelope tag')
        
        # for each envelope get the body
        for envelope in envelopes:
            body = find_child("Body", envelope)
            if not body:
                return Fault('WSManCLI', 
                         'Internal Server Error(WSManCLI Provider)', 
                         'Invalid response format - no Body tag in Envelope')
            
            # Aggregate items from both the EnumerateResponse and the PullResponse
            items = find_child("Items", find_child("EnumerateResponse", body))
            if items: 
                item_nodes.append(items)
            else:    
                items = find_child("Items", find_child("PullResponse", body))
                if items: 
                    item_nodes.append(items)
                else:
                    item_nodes.append(body)
            
        
        for item in item_nodes:
            # Decide if this is an Instance, Association or Reference, or Fault
            for child in item.get('children', []):
                qname = child.get('type', '')
                name = child.get('name', '')
                
                print name
                # Response Object
                response = None
                
                # End point reference (EPR)
                if name == 'EndpointReference':                
                    for child_ in child.get('children', []):
                        if child_.get('name', '') == 'ReferenceParameters':
                            response = self.response_from_reference(name, child_)
                
                
                elif name == "Fault" or name == "WSManFault":
                    return self.response_from_fault(child)
                
                elif name == "IdentifyResponse":
                    return self.response_from_identify(child)
                           
                # Instance or Association
                else:
                    response = Instance(name) if not self.is_association(child) else Association(name)                
                    
                    # Build the instance attributes
                    for child_ in child.get('children', []):
                        
                        # Get the name and value
                        key = child_.get('name', None)
                        value = child_.get('value', None)
                        
                        # Construct the association
                        if key and child_.get('children', []):                                                
                            for child__ in child_.get('children', []):
                                if child__.get('name', '') == 'ReferenceParameters':
                                    value = self.response_from_reference(key, child__)
                                elif child__.get('name', '') == 'EndpointReference': # Change in XML structure so that Reference Params and Endpoint References can occur at the same level
                                    for child___ in child__.get('children',[]):
                                        if child___.get('name', '') == 'ReferenceParameters':
                                            value = self.response_from_reference(key, child___)                                                                        
                        if key:                                
                            response.set(key, value)
                                                    
                if response:
                    responses.append(response)
                
        return responses
    
    def response_from_fault(self, node):
        """
        Construct the response object from the fault node dictionary
        
        @param node: XML fault dictionary
        @type node: Dictionary
        
        @return: Response object
        @rtype: L{Response}
        """
        
        code   = 'WSManCLI'
        reason = 'Internal Server Error (WSManCLI Provider)'
        detail = 'Internal Server Error (WSManCLI Provider)'
        
        log.debug ("Generating a fault from response %s" % node.get('name', ''))
        try:
            for child in node.get('children', []):
                name = child.get('name', '')
                if name == 'Code':
                    for x in child.get('children', []):
                        if x.get("name", '') == 'Subcode':
                            for y in x.get('children', []):
                                if y.get("name", '') == 'Value':
                                    code = y.get('value', '')                                                                    
                elif name == 'Reason':
                    for x in child.get('children', []):
                        if x.get("name", '') == 'Text':                        
                            reason = x.get('value', '')
                            
                elif name == 'Detail':
                    for x in child.get('children', []):
                        if x.get("name", '') == 'FaultDetail':                        
                            detail = x.get('value', '')
        except:
            log.warn("Error processing FAULT %s, %s" % (sys.exc_info()[0],sys.exc_info()[0]))
            pass
                        
        return Fault(code, reason, detail)        
    
    

        
    def get_response(self,node):
        """
        Gets a response object from parsed output.  Calls generate_response
        with a specific starting node.  This method tries to find where the 
        generate_response method should start processing. 
        
        @param node: dict representation of XML output
        @type node: dict
        
        @return: Response object
        @rtype: L{Response}
        """
        
        return self.generate_response(node)
        
    
    def parse(self, output):
        """
        Parse the output into one of the response formats.
        
        @param output: Output from the transport
        @type output: String
        """
        #print "---> Size of output", len(output)
        
        # Strip out the XML decl elements by pushing anything before the XML DECL and examining the tail
        output_frags = []
        tail = output
        found_decl = False
        
        while len(tail) > 0:
            
            start = tail.find("<?xml ")
            end = -1
            
            #print "SPLICE: len", len(tail), "start", start
            
            # a xml decl element was found
            if start > -1:
                end = tail.find("?>")
                
                if found_decl:
                    output_frags.append(tail[0:start])
                    
                found_decl = True
                tail = tail[end + 2 : ]
                
            else:
                # xml decl NOT found in the tail - append the tail to the output frags
                output_frags.append(tail)
                tail = ""
                
        output = " ".join(output_frags)
        
        #print "---> Size of spliced output", len(output)
        
        # put our own wrapper on around it
        output = "<climate:env>" + output + "</climate:env>"
        
        #print "\nParsing \n", output[0:40],"...",output[-20:],"\n"

        # Get the  dictionary representation of the extracted XML
        xml_dict = Parser().parse(output) if output else {}
        
        # Hold on to the body node - all responses have a body node
        return self.get_response(xml_dict)
        
        
        
    def remote_options(self, remote):
        """
        Generate the options for a remote command.
        
        @param remote: Remote configuration object
        @type remote: L{Remote}
        
        @return: Remote command line options
        @rtype: String
        """
        
        if not remote:
            return ''
        remote_  = '-u %s ' % remote.username
        remote_ += '-p %s ' % remote.password
        remote_ += '-h %s ' % remote.ip
        remote_ += '-P 443 -j utf-8 -y basic -V -v -c Dummy '
        return remote_
    
    
    def identify(self, remote=None, raw=False):
        """
        Identify WS-Man implementation
        
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        @return: L{Response} object or the raw XML response
        @rtype: L{Response}
        """    
        # Construct the command
        command = 'wsman identify -o -m 512 '
        command += self.remote_options(remote)
        
        # Use the transport and execute the command
        output = self.get_transport().execute(command)        
        if raw: 
            return output
        else:
            # Parse the output into a response object
            return self.parse(output)
    
    def enumerate(self, cim_class, cim_namespace, remote=None, raw=False, uri_host=""):
        """
        Enumerate the CIM class.
        
        @param cim_class: CIM class to be enumerated
        @type cim_class: String
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response object after enumeration
        @rtype: List of L{Response} objects/ L{Fault}
        """
        
        # Construct the command
        enumerate_command = 'wsman -o -m 512 '
        enumerate_command += self.remote_options(remote)
        enumerate_command += '-N %s enumerate %s/wbem/wscim/1/cim-schema/2/%s' % (cim_namespace, uri_host, cim_class)
        
        # Use the transport and execute the command
        output = self.get_transport().execute(enumerate_command)
        
        if raw:
            return output
        else:
            # Parse the output into a response object
            return self.parse(output)
    

    def enumerate_keys(self, cim_class, cim_namespace, remote=None, raw=False, uri_host=""):
        """
        Enumerate the keys for the cim class.
        
        @param cim_class: CIM class for key enumeration
        @type cim_class: String
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response objects after enumerating the keys
        @rtype: List of L{Reference} objects/ L{Fault}
        """
        
        # Construct the command
        enumerate_command = 'wsman -M epr -o -m 512 '
        enumerate_command += self.remote_options(remote)
        enumerate_command += '-N %s enumerate %s/wbem/wscim/1/cim-schema/2/%s' % (cim_namespace, uri_host, cim_class)
       
        print "Executing command %s" % enumerate_command
        # Use the transport and execute the command
        output = self.get_transport().execute(enumerate_command)
        
        if raw:
            return output
        else:
            # Parse the output into a response object
            return self.parse(output)
        
    
    def associators(self, reference, cim_namespace, remote=None, raw=False, uri_host=""):
        """
        Do a associators operation for an instance.
        
        @param reference: CIM instance response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response object after enumerating the keys
        @rtype: L{Instance}         
        """
         # Get the keys reference for this instance (This needs to be set prior to the get)
        if reference and\
              isinstance(reference,Reference):
                
            # Important: 
            # Form the query from the reference and not from the enumerated instance
            pairs = []
            for (key, value) in reference.items:
                if key != '__cimnamespace': # Ignore this key
                                        
                    # Assemble the intermediate value since it is stored as a list
                    # TODO: What if the value is an array? How do we assemble an array for a get query?
                    value_ = value[0] if value else ''
                    value_ = value_ if value_ else ''
                    
                    # If there are any spaces in the value then we need to single quote the string
                    safe_value = '\"%s\"' % value_ if value_.find(' ') > 0 else value_
                    pairs.append('%s=%s' % (key, safe_value))
            
            # Get the query parameter for command construction
            query = ','.join(pairs) + ' '
            
            # Construct the command
            get_command = 'wsman -o -m 512 '
            get_command += self.remote_options(remote)
            get_command += '-N %s associators %s/wbem/wscim/1/* ' % (uri_host, reference.get("__cimnamespace",[cim_namespace])[0])
            get_command += '--filter %s?%s --dialect http://schemas.dmtf.org/wbem/wsman/1/cimbinding/associationFilter' % (reference.resource_uri, query)
        
            print "Executing command %s" % get_command
            # Use the transport and execute the command
            output = self.get_transport().execute(get_command)
            
            if raw:
                return output
            else:
                # Parse the output into a response object
                return self.parse(output)
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
        
        # Generate a Fault
        return Fault('WSManCLI',\
                     'Get is not supported on this instance or the reference is not set or it returned an error.',\
                     'WinRM provider for WSMAN returned an error.')
    
    def references(self, reference, cim_namespace, remote=None, raw=False, uri_host=""):
        """
        Do a references operation for an instance.
        
        @param reference: CIM instance response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response object after enumerating the keys
        @rtype: L{Instance}         
        """
         # Get the keys reference for this instance (This needs to be set prior to the get)
        if reference and\
              isinstance(reference,Reference):
                
            # Important: 
            # Form the query from the reference and not from the enumerated instance
            pairs = []
            for (key, value) in reference.items:
                if key != '__cimnamespace': # Ignore this key
                                        
                    # Assemble the intermediate value since it is stored as a list
                    # TODO: What if the value is an array? How do we assemble an array for a get query?
                    value_ = value[0] if value else ''
                    value_ = value_ if value_ else ''
                    
                    # If there are any spaces in the value then we need to single quote the string
                    safe_value = '\"%s\"' % value_ if value_.find(' ') > 0 else value_
                    pairs.append('%s=%s' % (key, safe_value))
            
            # Get the query parameter for command construction
            query = ','.join(pairs) + ' '
            
            # Construct the command
            get_command = 'wsman -o -m 512 '
            get_command += self.remote_options(remote)
            get_command += '-N %s references %s/wbem/wscim/1/* ' % (uri_host, reference.get("__cimnamespace",[cim_namespace])[0])
            get_command += '--filter %s?%s --dialect http://schemas.dmtf.org/wbem/wsman/1/cimbinding/associationFilter' % (reference.resource_uri, query)
        
            print "Executing command %s" % get_command
            # Use the transport and execute the command
            output = self.get_transport().execute(get_command)
            
            if raw:
                return output
            else:
                # Parse the output into a response object
                return self.parse(output)
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
        
        # Generate a Fault
        return Fault('WSManCLI',\
                     'Get is not supported on this instance or the reference is not set or it returned an error.',\
                     'WinRM provider for WSMAN returned an error.')
        
    def set(self, reference, cim_namespace, remote=None, properties={}, raw=False):
        """
        Sets the properties of an instance using the properties argument.
        
        @param reference: CIM reference response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param properties: The properties and values to set
        @type properties: dict 
        
        @return: Instance object or Fault
        @rtype: L{Response}
        """
        # Get the keys reference for this instance (This needs to be set prior to the get)
        if reference and\
            isinstance(reference, Reference):
            
            # Important: 
            # Form the query from the reference and not from the enumerated instance
            pairs = []
            for (key, value) in reference.items:
                                        
                # Assemble the intermediate value since it is stored as a list
                # TODO: What if the value is an array? How do we assemble an array for a get query?
                value_ = value[0] if value else ''
                value_ = value_ if value_ else ''
                
                # JCT: used to wrap values with paces in double quotes - now we wrap the entire uri
                safe_value = value_
                pairs.append('%s=%s' % (key, safe_value))
            
            # Get the query parameter for command construction
            query = ','.join(pairs)
            classname = reference.classname
            
            # Construct the command
            get_command  = 'winrm put \"%s?%s\" ' % (reference.resource_uri, query)
            get_command += self.remote_options(remote)
            
            for k,v in properties.items():
                get_command += '-k \"%s=%s\" ' % (k,v)
            
            
            # Use the transport and execute the command
            output = self.get_transport().execute(get_command)
            if raw:
                return output
            else:
                return self.parse(output)
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
        
        # Generate a Fault
        return Fault('WSManCLI',\
                     'Get is not supported on this instance or the reference is not set or it returned an error.',\
                     'WinRM provider for WSMAN returned an error.')
        
        
    
    def invoke(self, reference, command, arguments, remote=None, raw=False):
        """
        Do a get operation for the instance
        
        @param reference: CIM reference response object
        @type reference: L{Reference}
        @param command: The command to invoke
        @type command: String
        @param arguments: A path to a file, or a dictionary of key/value paris used in the command
        @type arguments: String/dict
        @param remote: Remote configuration object
        @type remote: L{Remote}
        
        @return: Response object from the call, either a Reference, Fault, or an Result
        @rtype: L{Response}    
        """
        
        if reference and isinstance(reference, Reference):
            # Important: 
            # Form the query from the reference and not from the enumerated instance
            pairs = []
            
            for (key, value) in reference.items:
                # Assemble the intermediate value since it is stored as a list
                # TODO: What if the value is an array? How do we assemble an array for a get query?
                value_ = value[0] if value else ''
                value_ = value_ if value_ else ''
                
                # JCT: used to wrap values with paces in double quotes - now we wrap the entire uri
                safe_value = value_
                pairs.append('%s=\"%s\"' % (key, safe_value))
            
            # Get the query parameter for command construction
            query = ','.join(pairs)
            classname = reference.classname
            
            # Construct the command
            get_command  = 'wsman invoke \"%s?%s\" -a \"%s\" ' % (reference.resource_uri, query, command)
            get_command += self.remote_options(remote)
            
            # construct the options
            if isinstance(arguments, basestring):
                get_command += "--input=\"%s\"" % arguments
            elif isinstance(arguments, dict):
                for k,v in arguments.items():
                    get_command += '-k \"%s=%s\" ' % (k,v)
             
            output = self.get_transport().execute(get_command)
            
            if raw:
                return output
            else:
                instance_or_fault = self.parse(output)
                if not isinstance(instance_or_fault, Fault) and isinstance(instance_or_fault, list):                
                    if len(instance_or_fault) > 0:
                        return instance_or_fault[0]
                    
                return instance_or_fault
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
            
        # Generate a Fault
        return Fault('WSManCLI',\
                     'Get is not supported on this instance or the reference is not set or it returned an error.',\
                     'WinRM provider for WSMAN returned an error.')
        
    
    
    def get(self, reference, cim_namespace, remote=None, raw=False):
        """
        Do a get operation for an instance. The instance object should already have
        a key reference established before doing a get operation.
        
        @param reference: CIM instance response object
        @type reference: L{Reference}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        
        @return: Response object after enumerating the keys
        @rtype: L{Instance}         
        """
        
        # Note: There are no gets on an association instance
    
        # Get the keys reference for this instance (This needs to be set prior to the get)
        if reference and\
		      isinstance(reference,Reference):
                
            # Important: 
            # Form the query from the reference and not from the enumerated instance
            pairs = []
            for (key, value) in reference.items:
                if key != '__cimnamespace': # Ignore this key
                                        
                    # Assemble the intermediate value since it is stored as a list
                    # TODO: What if the value is an array? How do we assemble an array for a get query?
                    value_ = value[0] if value else ''
                    value_ = value_ if value_ else ''
                    
                    # If there are any spaces in the value then we need to single quote the string
                    safe_value = '\"%s\"' % value_ if value_.find(' ') > 0 else value_
                    pairs.append('%s=%s' % (key, safe_value))
            
            # Get the query parameter for command construction
            query = ','.join(pairs) + ' '
            
            # Construct the command
            get_command = 'wsman -o -m 512 '
            get_command += self.remote_options(remote)
            get_command += '-N %s get %s?%s' % (reference.get("__cimnamespace",[cim_namespace])[0], reference.resource_uri, query)
        
            print "Executing command %s" % get_command
            # Use the transport and execute the command
            output = self.get_transport().execute(get_command)
	        
            if raw:
                return output
            
            else:
                # Parse the output into a response object
                instance_or_fault = self.parse(output)
                if not isinstance(instance_or_fault, Fault) and isinstance(instance_or_fault, list):                
                    if len(instance_or_fault) > 0:
                        return instance_or_fault[0]
                    
                return instance_or_fault
        
        else:
            raise TypeError("reference argument must be of type Reference.  Use values from enumerate_keys instead?")
        
        # Generate a Fault
        return Fault('WSManCLI',\
                     'Get is not supported on this instance or the reference is not set or it returned an error.',\
                     'WinRM provider for WSMAN returned an error.')
        



"""
Process transport based on subprocess

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.


import os
import sys
import time
import logging
import subprocess
from .. import Transport

log = logging.getLogger("WSMAN.transport")

class Subprocess(Transport):
    """
    Subprocess based transport
    """
    
    def __init__(self):
        """
        Constructor for the CLI-mate transport
        """
        
        # Base class
        super(Subprocess, self).__init__()
        
    
    def shell(self):
        """
        Return the shell based on the operating system.
        
        @return: The shell command
        @rtype: String
        """
        
        if sys.platform == 'win32':
            return 'cmd.exe'        
        return 'sh'
    
    
    def delimiter(self):
        """
        Delimiter for the shell based on the operating system
        
        @return: The delimiter for the shell commands
        @rtype: String
        """
        
        if sys.platform == 'win32':
            return '\r\n'        
        return '\n'
    
    
    def execute(self, command):
        """
        Execute the command and return the output.
        
        @param command: The command constructed by the provider.
        @type command: String
        
        @return: The output from the command execution 
        @rtype: String
        """
        start = time.time() 
        process = subprocess.Popen(command,\
                                   stdout=subprocess.PIPE,\
                                   stderr=subprocess.PIPE,\
                                   shell=True)
        (stdout, stderr) = process.communicate()
        duration = time.time() - start        
        output =  stdout + stderr
        log.info("Command Completed in %0.3f s" % duration, extra={'command': command, 'output': output, 'duration':duration})
        return output
        



"""
Base WSMan provider

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

class WSManProvider(object):
    """
    WS-Management provider
    """
    
    
    def __init__(self, transport):
        """
        Constructor for the WSMan provider class.
        
        @param transport: Transport object to execute the constructed command
        @type transport: L{Transport}
        """
        
        # Transport for the provider
        self.__transport = transport
        
        
    def get_transport(self):
        """
        Get the transport for this provider.
        
        @return: Transport for the provider
        @rtype: L{Transport}
        """
        
        return self.__transport
    
    def identify(self, remote=None):
        """
        Identify WS-Man implementation
        
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param raw: Determines if the method should return the XML output from the transport, or a L{Response} object.
                    If you want to do your own parsing of the XML output, then set this parameter to True. (default=False)
        @type raw: bool
        @return: L{Response} object or the raw XML response
        @rtype: L{Response}
        """
        raise NotImplementedError("This method needs to be implemented in the derived class.")
    
    
    def enumerate(self, cim_class, cim_namespace, remote=None, raw=False, uri_host=""):
        """
        Enumerate the cim class.
        
        @param cim_class: CIM class to be enumerated
        @type cim_class: String
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response object after enumeration
        @rtype: List of L{Response} objects/ L{Fault}
        """
        
        raise NotImplementedError("This method needs to be implemented in the derived class.")
    
    
    def enumerate_keys(self, cim_class, cim_namespace, remote=None, uri_host=""):
        """        
        Enumerate the keys for the cim class.
        
        @param cim_class: CIM class for key enumeration
        @type cim_class: String
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response object after enumerating the keys
        @rtype: L{Reference}
        """
        
        raise NotImplementedError("This method needs to be implemented in the derived class.")
    
    def associators(self, instance, cim_namespace, remote=None, uri_host=""):
        """
        Do an associators operation for the instance
        
        @param instance: CIM instance response object
        @type instance: L{Instance}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response object after enumerating the keys
        @rtype: L{Instance}         
        """
        
        raise NotImplementedError("This method needs to be implemented in the derived class.")
    
    def references(self, instance, cim_namespace, remote=None, uri_host=""):
        """
        Do a references operation for the instance
        
        @param instance: CIM instance response object
        @type instance: L{Instance}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        @param uri_host: The host portion of the resource URI
        @type uri_host: L{String}
        
        
        @return: Response object after enumerating the keys
        @rtype: L{Instance}         
        """
        
        raise NotImplementedError("This method needs to be implemented in the derived class.")
    
    def get(self, instance, cim_namespace, remote=None):
        """
        Do a get operation for the instance
        
        @param instance: CIM instance response object
        @type instance: L{Instance}
        @param cim_namespace: Namespace of the CIM class
        @type cim_namespace: String
        @param remote: Remote configuration object
        @type remote: L{Remote}
        
        @return: Response object after enumerating the keys
        @rtype: L{Instance}         
        """
        
        raise NotImplementedError("This method needs to be implemented in the derived class.")
    
    
    def invoke(self, reference, command, arguments, remote=None):
        """
        Do a get operation for the instance
        
        @param reference: CIM reference response object
        @type reference: L{Reference}
        @param command: The command to invoke
        @type command: String
        @param arguments: A path to a file, or a dictionary of key/value paris used in the command
        @type arguments: String/dict
        @param remote: Remote configuration object
        @type remote: L{Remote}
        
        @return: Response object from the call, either a Reference, Fault, or an Result
        @rtype: L{Response}    
        """
        
        raise NotImplementedError("This method needs to be implemented in the derived class.")
    


"""
WS-Management provider

@copyright: 2010-2012
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

import sys
from winrm import WinRM
from wsmancli import WSManCLI

# Factory for the WSMan Provider
class WSManProviderFactory(object):
    """
    Factory for the WSMan provider
    """
    
    def __init__(self, transport):
        """
        Initialize the factory with the transport.
        
        @param transport: Transport object to execute the constructed command
        @type transport: L{Transport}
        """
        
        # Set the transport
        self.__transport = transport    
    
    
    def get_provider(self):
        """
        Get the provider based on the specified criterion.
        
        @return: WSManProvider that matches the specified criterion.
        @rtype: L{WSManProvider}
        """
        
        if sys.platform == 'win32':
            return WinRM(self.__transport)
        
        return WSManCLI(self.__transport) 




"""
Instance response

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@author: Vijay Halaharvi <vijay_halaharvi@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.

from . import KeyResponse
from ..mixin.dictionary import DictionaryMixin


class Instance(KeyResponse, DictionaryMixin):
    """
    Instance response object
    """
    
    def __init__(self, name):
        """
        Constructor for the Instance class
        
        @param name: Name of the instance
        @type name: String
        """
                
        # Call the base class
        super(Instance, self).__init__()
        
        # Store the name of the class that this is an instance of. 
        self.__name = name
    
    def toString(self):
        """
        show the object in a reader friendly format
        """
        strs = [self.name]
        
        for k,v in self.items:
            #strs.append("\t%s = %s" % (k,','.join(v) if v else "" ))
            strs.append("\t%s=%s" % (k, ",".join(map(lambda x: x if x else "" ,v)) ))
        return "\n".join(strs)
        
        
    # Properties of this class    
    name = property(fget=lambda x: x.__name)
        


"""
WSMan Commmand Log Handler Module

@copyright: 2010-2012
@author: Joseph Tallieu <joseph_tallieu@dell.com>
@organization: Dell Inc. - PG Validation
@license: GNU LGLP v2.1
"""
#    This file is part of WSManAPI.
#
#    WSManAPI is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 2.1 of the License, or
#    (at your option) any later version.
#
#    WSManAPI is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with WSManAPI.  If not, see <http://www.gnu.org/licenses/>.